#include "land.h"
#include "vector.h"


Landscape::Landscape(int s, int sB)
{
	size = s;
	sizeBits = sB;

	numfaces = size * size;
	numverts = (size+1) * (size+1);

	verts = new vertexData[numverts];
	faces = new faceData[numfaces];

	clear();

	char filename[30];
	for (int x = 0; x < TEX_LAND_COUNT; x++)
	{
		sprintf(filename,"tex256/land%d",x);
		landtex[x] = new Texture(filename,1);
	}
	for (x = 0; x < TEX_TRANS_COUNT; x++)
	{
		sprintf(filename,"tex64/trans%d",x);
		transtex[x] = new Texture(filename,1);
	}

}


Landscape::~Landscape()
{
	delete[] verts;
	delete[] faces;

	for (int x = 0; x < TEX_LAND_COUNT; x++)
	{
		delete landtex[x];
	}
	for (x = 0; x < TEX_TRANS_COUNT; x++)
	{
		delete transtex[x];
	}
}

void Landscape::clear()
{
	for (int x = 0; x < numverts; x++)
	{
		verts[x].height = 0;
		verts[x].shade = 0;
	}
	for (x = 0; x < numfaces; x++)
	{
		faces[x].landtype = -1;
		faces[x].landtex = 0;
		faces[x].transtex1 = -1;
		faces[x].transtex2 = -1;
	}
}

float parabolicInterpolate(float v1, float v2, float blend)
{
	if (blend < 0) blend = 0.0;
	if (blend > 1) blend = 1.0;	

	float sinval = float( sin( PI * (1.5 + blend/2) ) + 1 );

	return ( v1 * (1-sinval)  +  v2 * sinval);
}

float cubicInterpolate(float v1, float v2, float blend)
{
	if (blend < 0) blend = 0.0;
	if (blend > 1) blend = 1.0;	

	if (v1 > v2)
	{
		blend = float(PI * (blend + 0.5));
		double sinval = (sin(blend)+1) / float(2);
		return float(v2 + (sinval * (v1-v2)));	
	}
	else
	{
		if (blend < 0.5)
		{
			blend = float(PI * (blend + 1.5));
		}
		else
		{
			blend = float(PI * (blend - 0.5));
		}
		double sinval = (sin(blend)+1) / float(2);
		return float(v1 + (sinval * (v2-v1)));	
	}
}

void Landscape::generateHeights(int wavelength, float amp)
{
	if (amp <= 1.0) amp = 1.0;

	float* heights = new float[(size+1)*(size+1)];
	float* finals = new float[(size+1)*(size+1)];

	for (int i = 0; i < (size+1)*(size+1); i++)
	{
		finals[i] = -1.0;
	}

	for (int x = 0; x <= size; x += wavelength)
	{
		for (int y = 0; y <= size; y += wavelength)
		{
			heights[ (y * (size+1)) + x ] = parabolicInterpolate(0,amp,float(rand() % int(amp))/amp);	
		}
	}

	for (x = 0; x < size; x += wavelength)
	{
		for (int y = 0; y < size; y += wavelength)
		{

			float p1 = heights [ (y * (size+1)) + x ];
			float p2 = heights [ (y * (size+1)) + x + wavelength];
			float p3 = heights [ ((y+wavelength) * (size+1)) + x + wavelength];
			float p4 = heights [ ((y+wavelength) * (size+1)) + x ];

			for (int gridx = x; gridx <= x+wavelength; gridx++)
			{
				for (int gridy = y; gridy <= y+wavelength; gridy++)
				{
					float blendx = float(gridx - x) / float(wavelength);
					float blendy = float(gridy - y) / float(wavelength);

					float height1 = cubicInterpolate(p1,p2,blendx);
					float height2 = cubicInterpolate(p4,p3,blendx);

					
					if ( finals [ (gridy * (size+1)) + gridx ] < 0)
					{	
						finals [ (gridy * (size+1)) + gridx ] = cubicInterpolate(height1,height2,blendy);
					}
				}
			}

		}
	}

	for (int a = 0; a <= size; a++)
	{
		for (int b = 0; b <= size; b++)
		{
			vert(a,b)->height += finals [ (b * (size+1)) + a ];
		}
	}

	delete[] heights;
	delete[] finals;

	if (wavelength >= 16) generateHeights(wavelength / 2, amp / 2);
}

void Landscape::generateShades()
{
	for (int i = 0; i <	numverts; i++)
	{
		verts[i].shade = 0;
	}

	float sunvect[3] = { 3,3,6 };
	normalize(sunvect);

	for (int x = 1; x < size; x++)
	{
		for (int y = 1; y < size; y++)
		{			
			float v1[3] = { 1, 0, vert(x+1,y)->height - vert(x,y)->height };
			float v2[3] = { 0, 1, vert(x,y+1)->height - vert(x,y)->height };

			float norm[3];
			cross(v1,v2,norm);
			normalize(norm);

			float shad[3];
			cross(sunvect,norm,shad);

			vert(x,y)->shade = 1 - vlength(shad);
		}
	}
}

void Landscape::smoothHeightEdges(int bufferSize)
{
	for (int x = 0; x <= size; x++)
	{
		for (int y = 0; y <= size; y++)
		{
			
		}
	}
}

void Landscape::generate()
{
	
	generateHeights(size/2, 15);

	smoothHeightEdges(45);

	generateShades();

}


void Landscape::drawMesh()
{
	
	for (int x = 0; x < size; x++)
	{
		for (int y = 0; y < size; y++)
		{
			glBegin(GL_LINE_LOOP);
			
			glVertex3f((float)x,(float)y,vert(x,y)->height);
			glVertex3f((float)x+1,(float)y,vert(x+1,y)->height);
			glVertex3f((float)x+1,(float)y+1,vert(x+1,y+1)->height);
			glVertex3f((float)x,(float)y+1,vert(x,y+1)->height);

			glEnd();
		}
	}
}

void Landscape::drawPlainSolid()
{
	glColor3f(1,1,1);

	for (int x = 0; x < size; x++)
	{
		for (int y = 0; y < size; y++)
		{
			float tx = float(x) / float(4);
			float ty = float(y) / float(4);

			int tex = int(vert(x,y)->height / float(3));
			if (tex > TEX_LAND_COUNT) tex = 6;
			landtex[0]->enable();

			glBegin(GL_QUADS);
			
			glColor4f(vert(x,y)->shade,vert(x,y)->shade,vert(x,y)->shade,0);
			glTexCoord2f(tx,ty);
			glVertex3f((float)x,(float)y,vert(x,y)->height);

			glColor4f(vert(x+1,y)->shade,vert(x+1,y)->shade,vert(x+1,y)->shade,0);
			glTexCoord2f(tx+float(1)/float(4),ty);
			glVertex3f((float)x+1,(float)y,vert(x+1,y)->height);

			glColor4f(vert(x+1,y+1)->shade,vert(x+1,y+1)->shade,vert(x+1,y+1)->shade,0);
			glTexCoord2f(tx+float(1)/float(4),ty+float(1)/float(4));
			glVertex3f((float)x+1,(float)y+1,vert(x+1,y+1)->height);

			glColor4f(vert(x,y+1)->shade,vert(x,y+1)->shade,vert(x,y+1)->shade,0);
			glTexCoord2f(tx,ty+float(1)/float(4));
			glVertex3f((float)x,(float)y+1,vert(x,y+1)->height);

			glEnd();
			
			transtex[2]->enable();

			glEnable(GL_BLEND);
			glBlendFunc(GL_SRC_ALPHA, GL_ONE);

			/*glBegin(GL_QUADS);
			
			glColor4f(0,0,0,1);
			glTexCoord2f(tx,ty);
			glVertex3f((float)x,(float)y,vert(x,y)->height);

			
			glTexCoord2f(tx+float(1)/float(4),ty);
			glVertex3f((float)x+1,(float)y,vert(x+1,y)->height);

			
			glTexCoord2f(tx+float(1)/float(4),ty+float(1)/float(4));
			glVertex3f((float)x+1,(float)y+1,vert(x+1,y+1)->height);

			
			glTexCoord2f(tx,ty+float(1)/float(4));
			glVertex3f((float)x,(float)y+1,vert(x,y+1)->height);

			glEnd();*/

			landtex[2]->enable();
			
			glBlendFunc(GL_ONE, GL_DST_ALPHA);

			/*glBegin(GL_QUADS);
			
			glColor3f(vert(x,y)->shade,vert(x,y)->shade,vert(x,y)->shade);
			glTexCoord2f(tx,ty);
			glVertex3f((float)x,(float)y,vert(x,y)->height);

			glColor3f(vert(x+1,y)->shade,vert(x+1,y)->shade,vert(x+1,y)->shade);
			glTexCoord2f(tx+float(1)/float(4),ty);
			glVertex3f((float)x+1,(float)y,vert(x+1,y)->height);

			glColor3f(vert(x+1,y+1)->shade,vert(x+1,y+1)->shade,vert(x+1,y+1)->shade);
			glTexCoord2f(tx+float(1)/float(4),ty+float(1)/float(4));
			glVertex3f((float)x+1,(float)y+1,vert(x+1,y+1)->height);

			glColor3f(vert(x,y+1)->shade,vert(x,y+1)->shade,vert(x,y+1)->shade);
			glTexCoord2f(tx,ty+float(1)/float(4));
			glVertex3f((float)x,(float)y+1,vert(x,y+1)->height);

			glEnd();*/
			
			glDisable(GL_BLEND);

			
			glDisable(GL_DEPTH);

			glColor3f(1,1,1);

			/*glBegin(GL_LINES);

			
			glVertex3f((float)x,(float)y,vert(x,y)->height);
			glVertex3f((float)x+1,(float)y,vert(x+1,y)->height);
			glVertex3f((float)x+1,(float)y+1,vert(x+1,y+1)->height);
			glVertex3f((float)x,(float)y+1,vert(x,y+1)->height);
				

			glEnd();*/

			glEnable(GL_DEPTH);
			
		}
	}		
}

void Landscape::drawRegion(int x, int y, int s)
{
	FILE* file = fopen("debug.txt","w");
	
	fprintf(file,"%d %d %d",x,y,s);

	fclose(file);

	int fx = x+s;
	int fy = y+s;
	float tx = float(x * TEX_STRETCH_RATIO);
	float ty = float(y * TEX_STRETCH_RATIO);
	for (int i = x; i < fx; i++)
	{
		for (int j = y; j < fy; j++)
		{
			float shade1 = vert(i,j)->shade;
			float shade2 = vert(i+1,j)->shade;
			float shade3 = vert(i+1,j+1)->shade;
			float shade4 = vert(i,j+1)->shade;

			float height1 = vert(i,j)->height;
			float height2 = vert(i+1,j)->height;
			float height3 = vert(i+1,j+1)->height;
			float height4 = vert(i,j+1)->height;

			landtex[face(i,j)->landtex]->enable();

			glBegin(GL_QUADS);

				glColor3f(shade1,shade1,shade1);
				glTexCoord2f(tx,ty);
				glVertex3f((float)i, (float)j, height1);

				glColor3f(shade2,shade2,shade2);
				glTexCoord2f(tx+(float)TEX_STRETCH_RATIO,ty);
				glVertex3f((float)i+1, (float)j, height2);

				glColor3f(shade3,shade3,shade3);
				glTexCoord2f(tx+(float)TEX_STRETCH_RATIO,ty+(float)TEX_STRETCH_RATIO);
				glVertex3f((float)i+1, (float)j+1, height3);

				glColor3f(shade4,shade4,shade4);
				glTexCoord2f(tx,ty+(float)TEX_STRETCH_RATIO);
				glVertex3f((float)i, (float)j+1, height4);

			glEnd();
			ty += TEX_STRETCH_RATIO;

		}
		tx += TEX_STRETCH_RATIO;
		ty = float(y * TEX_STRETCH_RATIO);
	}
}

void Landscape::clipRegion(int x, int y, int s)
{
	if (s <= LAND_CLIP_LIMIT)
	{
		drawRegion(x,y,s);
		return;
	}
	
	int s2 = s / 2;
	int x2 = (x+s)/2;
	int x3 = x+s;
	int y2 = (y+s)/2;
	int y3 = y+s;

	double winx,winy,winz;

	int clip[9] = {0,0,0,0,0,0,0,0,0};

	if (gluProject(x,y,10,modelMatrix,projMatrix,viewport,&winx,&winy,&winz)) if (winx > 0 && winx < resx && winy > 0 && winy < resy) clip[0] = 1;
	if (gluProject(x2,y,10,modelMatrix,projMatrix,viewport,&winx,&winy,&winz)) if (winx > 0 && winx < resx && winy > 0 && winy < resy) clip[1] = 1;
	if (gluProject(x3,y,10,modelMatrix,projMatrix,viewport,&winx,&winy,&winz)) if (winx > 0 && winx < resx && winy > 0 && winy < resy) clip[2] = 1;
	if (gluProject(x,y2,10,modelMatrix,projMatrix,viewport,&winx,&winy,&winz)) if (winx > 0 && winx < resx && winy > 0 && winy < resy) clip[3] = 1;
	if (gluProject(x2,y2,10,modelMatrix,projMatrix,viewport,&winx,&winy,&winz)) if (winx > 0 && winx < resx && winy > 0 && winy < resy) clip[4] = 1;
	if (gluProject(x3,y2,10,modelMatrix,projMatrix,viewport,&winx,&winy,&winz)) if (winx > 0 && winx < resx && winy > 0 && winy < resy) clip[5] = 1;
	if (gluProject(x,y3,10,modelMatrix,projMatrix,viewport,&winx,&winy,&winz)) if (winx > 0 && winx < resx && winy > 0 && winy < resy) clip[6] = 1;
	if (gluProject(x2,y3,10,modelMatrix,projMatrix,viewport,&winx,&winy,&winz)) if (winx > 0 && winx < resx && winy > 0 && winy < resy) clip[7] = 1;
	if (gluProject(x3,y3,10,modelMatrix,projMatrix,viewport,&winx,&winy,&winz)) if (winx > 0 && winx < resx && winy > 0 && winy < resy) clip[8] = 1;

	if (clip[0] || clip[1] || clip[3] || clip[4]) clipRegion(x,y,s2);
	if (clip[1] || clip[2] || clip[4] || clip[5]) clipRegion(x2,y,s2);
	if (clip[3] || clip[4] || clip[6] || clip[7]) clipRegion(x,y2,s2);
	if (clip[4] || clip[5] || clip[7] || clip[8]) clipRegion(x2,y2,s2);
}

void Landscape::drawLand()
{
	clipRegion(0,0,size);
	glDisable(GL_TEXTURE_2D);
}


vertexData* Landscape::vert(int x, int y)
{
	return &verts[ (y << sizeBits) + y + x ];
}

faceData* Landscape::face(int x, int y)
{
	return &faces[ (y << sizeBits) + x ];
}

void Landscape::updateViewMatricies()
{
	glGetDoublev(GL_MODELVIEW_MATRIX,modelMatrix);
	glGetDoublev(GL_PROJECTION_MATRIX,projMatrix);
	glGetIntegerv(GL_VIEWPORT,viewport);
}